local _, class = UnitClass("player")
if class ~= "HUNTER" then return end

local emod = clcInfo.env
local debug = clcInfo.debug
local db

-- status
local s_ctime, s_realtime, s_delta, s_beast = 0, 0, 0, 0

-- focus storage
local focus

-- mode, 0 is Careful Aim, 1 is Standard, 2 is Kill Shot
local mode

-- hunter instant shots incurr a 1s GCD
emod.H_INSTANT = 1.0

emod.spells = {
	["Serpent Sting"] = 1978,
	["Arcane Shot"] = 3044,
	["Steady Shot"] = 56641,
	["Chimera Shot"] = 53209,
	["Dire Beast"] = 120679,
	["Kill Shot"] = 53351,
	["Rapid Fire"] = 3045,
	["Aimed Shot"] = 19434,
	["Arcane Shot"] = 3044,
	["A Murder of Crows"] = 131894,
	["Auto Shot"] = 75,
	["Cobra Shot"] = 77767,
	["Bestial Wrath"] = 19574,
	["Kill Command"] = 34026,
	["Stampede"] = 121818,
	["Glaive Toss"] = 117050,
	["Explosive Trap"] = 13813,
	["Focus Fire"] = 82692,
	["Lynx Rush"] = 120697,
	["Black Arrow"] = 3674,
	["Explosive Shot"] = 53301,
	["Multi-Shot"] = 2643,
	["Viper Venom"] = 118976,
	["Serpent Spread"] = 87935,
}

emod.auras = {
	["Serpent Sting"] = GetSpellInfo(1978),
	["Steady Focus"] = GetSpellInfo(53224),
	["Fire!"] = GetSpellInfo(82926),
	["A Murder of Crows"] = GetSpellInfo(131894),
	["Rapid Fire"] = GetSpellInfo(3045),
	["Frenzy"] = GetSpellInfo(19615),
	["The Beast Within"] = GetSpellInfo(34692),
	["Lock and Load"] = GetSpellInfo(56453),
	["Black Arrow"] = GetSpellInfo(3674),
	["Cobra Shot"] = GetSpellInfo(77767) -- not really an aura but need the name localised for cast detection
}

emod.talents = {}
emod.recent = {
	[120679] = 0, -- dire beast
	[1978] = 0, -- serpent sting
	[2643] = 0, -- multi-shot
}

-- spell tracking
local recent_ss = 0

local spellframe = CreateFrame("Frame", nil, UIParent)
spellframe:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
spellframe:RegisterEvent("PLAYER_TALENT_UPDATE")
spellframe:RegisterEvent("PLAYER_ENTERED_WORLD")
spellframe:RegisterEvent("ZONE_CHANGED")
spellframe:SetScript("OnEvent", function(self, event, unit, name, rank, line, id)
	if event == "UNIT_SPELLCAST_SUCCEEDED" and unit == "player" then
		-- record the last time of desired shots in recent
		if emod.recent[id] ~= nil then
			emod.recent[id] = GetTime()
		end

		-- timing last cast steady shot, dire beast or resetting recent steady shots
		if id == emod.spells["Steady Shot"] then recent_ss = recent_ss + 1
		elseif id ~= emod.spells["Auto Shot"] then recent_ss = 0 end
		if id == emod.spells["Dire Beast"] then beast = GetTime() end
	elseif event == "PLAYER_ENTERED_WORLD" or event == "PLAYER_TALENT_UPDATE" then
		-- we're caching talents in emod.talents, so need to hook them when we log in
		-- or when we change spec
		emod:UpdateTalents()
	end
end)

-- gets the time from WoW and reset the value of s_realtime
function emod:GetTime()
	s_ctime = GetTime()

	s_realtime = 1
	s_delta = 0

	-- pulling us to realtime should also update the current
	-- focus
	focus = UnitPower("player")

	return s_ctime
end

-- advances the time to a given point and returns that value
-- sets s_realtime so we know to use cached values until
-- we're asked to return to normal time
function emod:SetTime(delta)
	s_delta = s_delta + delta
	s_ctime = s_ctime + delta
	s_realtime = 0
	return s_ctime
end

-- updates focus with a estimate of how much would have been regenerated
-- during the advanced period, along with how many attacks dire beast
-- should have had
function emod:UpdateFocus(ofocus, delta)
	local focus = ofocus
	local max = UnitPowerMax("player")

	-- now calculate how much generic regen we have made
	local __, active = GetPowerRegen()
	focus = min(max, focus + (active * delta))

	-- if dire beast is active we have to guess at it's possible regen
	if beast and beast > 0 and (GetTime() - beast) < 15.0 then
		local s_haste = 1 + GetMeleeHaste("player") / 100
		-- DB has base 2s attack speed, adjusted by haste
		local regen = (15 / (2.0 * s_haste * 5))
		focus = min(max, focus + (regen * delta))
	end

	-- if we know Viper Venom (ie. we're survival) then we need to account
	-- for serpent sting being on our primary target (can't really check others)
	if IsSpellKnown(emod.spells["Viper Venom"]) then
		if emod:GetTargetDebuff(emod.auras["Serpent Sting"]) > 0 then
			-- this is an approximation, to do this properly I would need to
			-- track the last known proc of viper venom, along with the last
			-- known proc of serpent sting and work out if we're going to hit
			-- a proc within the delta period.
			focus = min(max, focus + 1.0)
		end
	end

	return focus
end

function emod:GetFocus()
	return focus
end

function emod:SetFocus(delta)
	focus = emod:UpdateFocus(focus, delta)
end

function emod:UseFocus(f)
	-- if we have a time for beast within then we'll halve the
	-- cost of the focus. unfortunate that this needs to be hardcoded
	-- her really.
	if emod.s_beast and emod.s_beast > 0 then
		focus = focus * 0.5
	end

	focus = min(UnitPowerMax("player"), max(0, focus - f))
end

function emod:UpdateTalents()
	for i=1,GetNumTalents() do
		local __, __, __, __, selected, __ = GetTalentInfo(i)
		emod.talents[i] = selected
	end
end

function emod:GetGlobalCooldown()
	-- gcd
	local s_gcd
	local start, duration = GetSpellCooldown(emod.spells["Arcane Shot"])
	s_gcd = start + duration - GetTime()
	if s_gcd < 0 then s_gcd = 0 end
	return s_gcd
end

function emod:GetCooldown(id)
	-- returns 100 if we don't actually know the spell
	if not IsSpellKnown(id) then return 100 end

	local start, duration = GetSpellCooldown(id)
	local cd = start + duration - s_ctime - self:GetGlobalCooldown()
	if cd < 0 then return 0 end
	return cd
end

function emod:GetTimeUntilFocus(id)
	-- return the number of seconds we estimate we need to regen enough
	-- focus to use this particular spell
	local __, combatregen = GetPowerRegen()
	local __, __, __, cost = GetSpellInfo(id)
	if cost == 0 then return 0 end

	local current = UnitPower("player")
	if current >= cost then return 0 end

	if beast and beast > 0 and (GetTime() - beast) < 15.0 then
		local s_haste = 1 + GetMeleeHaste("player") / 100
		combatregen = combatregen + (15 / (2.0 * s_haste * 5))
	end

	if IsSpellKnown(emod.spells["Viper Venom"]) then
		if emod:GetTargetDebuff(emod.auras["Serpent Sting"]) > 0 then
			-- serpent sting adds a 3f/3s regen to base regen that does
			-- not factor in to the GetPowerRegen result
			combatregen = combatregen + 1.0 -- 3f / 3s
		end
	end

	return (cost - current) / combatregen
end

function emod:GetCastTime(id)
	local __, __, __, __, __, __, time = GetSpellInfo(id)
	return time / 1000
end

function emod:GetRecentSteadyShotCount()
	return recent_ss
end

function emod:GetRecentSerpentSting()
	local sting = emod.recent[emod.spells["Serpent Sting"]]
	local multi = emod.recent[emod.spells["Multi-Shot"]]

	if sting == 0 and multi == 0 then return false end

	-- had recent serpent sting fired, give 2s to allow for travel time
	if GetTime() - sting < 2.0 then
		return true
	end

	-- if survival (serpent spread) also include multishot
	if IsSpellKnown(emod.spells["Serpent Spread"]) then
		if GetTime() - multi < 2.0 then
			return true
		end
	end

	return false
end

function emod:GetTargetDebuff(debuff)
	local left = 0
	local name, __, __, __, __, __, expires, __, __, __, __ = UnitAura("target", debuff, nil, "PLAYER|HARMFUL")
	if expires and expires > 0 then
		left = max(0, expires - s_ctime)
	elseif name then
		-- debuffs without times (murder??!)
		left = 100
	end
	return left
end
