local _, class = UnitClass("player")
if class ~= "HUNTER" then return end

local modName = "__beastmastery"
local mod = clcInfo:RegisterClassModule(modName)
local emod = clcInfo.env
local debug = clcInfo.debug
local db

-- status
local s1, s2

local prio = {
	"se", -- serpent sting
	"bw", -- bestial wrath
	"rf", -- rapid fire
	"kc", -- kill command
	"ff", -- focus fire
	"stm", -- stampede
	"moc", -- murder of crows (talent)
	"db", -- dire beast
	"lr", -- lynx rush (talent)
	"gt", -- glaive toss (talent)
	"ks", -- kill shot
	"ac", -- arcane shot
	"cs", -- cobra shot
}

local function GetStatus()
	-- also updates focus
	s_ctime = emod:GetTime()

	-- gcd
	local start, duration = GetSpellCooldown(emod.spells["Arcane Shot"])
	local gcd = start + duration - s_ctime
	if gcd < 0 then gcd = 0 end

	-- we actually want to advance the model past the current GCD
	-- so that we make predictions based on the next available time
	-- that a shot can be used
	if gcd > 0 then
		s_ctime = emod:SetTime(gcd)
		emod:SetFocus(gcd)
	end

	-- update info
	emod.s_ss = emod:GetTargetDebuff(emod.auras["Serpent Sting"])
	emod.s_moc = emod:GetTargetDebuff(emod.auras["A Murder of Crows"])
	emod.s_sf = emod:GetBuff(emod.auras["Steady Focus"])
	emod.s_rf = emod:GetBuff(emod.auras["Rapid Fire"])
	emod.s_bw = emod:GetBuff(emod.auras["The Beast Within"])
	emod.s_frenzy = emod:GetBuffStacks(emod.auras["Frenzy"])

	-- fiddle serpent sting for inflight shots
	if emod.s_ss == 0 and emod:GetRecentSerpentSting() then
		emod.s_ss = 15.0
	end

	-- take a guess at whether we're facing a boss
	emod.s_boss = emod:UnitIsBoss("target")
end

local function BeastRotation()
	s1 = nil
	GetStatus()
	if debug.enabled then
		debug:Clear()
		debug:AddBoth("mode", mode)
		debug:AddBoth("s_ss", s_ss)
		debug:AddBoth("s_mm", s_mm)
		debug:AddBoth("s_sf", s_sf)
		debug:AddBoth("focus", focus)
		debug:AddBoth("s_moc", s_moc)
		debug:AddBoth("recent_ss", recent_ss)
	end

	emod:SetLastAbility(nil)

	local action
	s1, action = emod:GetNextAbility(prio)
	if debug.enabled then
		debug:AddBoth("s1", action)
		debug:AddBoth("s1f", emod:GetTimeUntilFocus(s1))
	end

	-- in order to try and stabilise the predictions of next events
	-- when casting cobra strike, if we're casting it then it should
	-- be forced to be the current action
	local spellname = UnitCastingInfo("player")
	if spellname == emod.auras["Cobra Shot"] then
		s1, action = emod:GetAbility("_cs").id, "_cs"
	end

	emod:SetLastAbility(action)

	s_otime = s_ctime
	-- each action needs to update s_ctime for the time taken to cast
	emod:GetAbility(action).UpdateStatus()

	s_otime = s_ctime - s_otime
	emod.s_ss = max(0, emod.s_ss - s_otime)

	-- now calculate how much generic regen we have made
	emod:SetFocus(s_otime)

	if debug.enabled then
		debug:AddBoth("s_ss", emod.s_ss)
		debug:AddBoth("s_mm", emod.s_mm)
		debug:AddBoth("focus", emod.focus)
	end

	s2, action = emod:GetNextAbility(prio)
	if debug.enabled then
		debug:AddBoth("s2", action)
	end
end

function emod.IconBeast1(...)
	BeastRotation(...)
	if secondarySkill then secondarySkill:DoUpdate() end
	return emod.IconSpell(s1, true)
end

local function SecondaryExec()
	return emod.IconSpell(s2, true)
end

local function ExecCleanup2()
	secondarySkill = nil
end

function emod.IconBeast2(...)
	secondarySkill = emod.___e
	secondarySkill:SetScript("OnUpdate", nil)
	secondarySkill.exec = SecondaryExec
	secondarySkill.ExecCleanup = ExecCleanup2
end
